#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/sched.h> 
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <linux/ioport.h>
#include <linux/slab.h>     
#include <linux/poll.h>     
#include <linux/proc_fs.h>
#include <linux/irq.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>

#include <asm/system.h>     
#include <asm/uaccess.h>
#include <config/nr/irqs.h>
#include <asm/irq.h>
#include <asm/ioctl.h>
#include <asm/unistd.h>
#include <asm/io.h>

static int major   = 240;
static struct pci_dev *PCI = NULL;


#define		DEV_NAME				"PCIE_DRV"
#define		SPCI_PRIVATE_NUMBER		0x1234

#define		TXDMA_SIZE				2048

static struct pci_device_id spci_tbl[] = {
	{0x1172, 0x0004, PCI_ANY_ID, PCI_ANY_ID, 0, 0, SPCI_PRIVATE_NUMBER },
	{0,}
};

MODULE_DEVICE_TABLE (pci, spci_tbl);

typedef struct { 
	
	unsigned long mmio_base[3];	
	
	void *mmio_vir[3];

	int offset[3];

	u_int *tx_vir;
	dma_addr_t txdma_phy;
	int dma_ch;
} spci_info_t;

#define PRINTK( fmt, args...)   printk("[kernel]%s:%d) "fmt, __FUNCTION__,__LINE__, ## args )

static int spci_open( struct inode *inode, struct file *filp )
{
	int	rtn;
	
	if ( PCI == NULL )
	{
		PRINTK( "not found PCI resource\n" );
		return -ENOMEM;
	}

	rtn = request_irq( PCI->irq, (void *)spci_interrupt, IRQF_SHARED, DEV_NAME, PCI );
	if ( 0 > rtn )
	{
		PRINTK("unable open IRQ (%d)\n", PCI->irq );
		return rtn;
	}
	PRINTK("IRQ num=%d\n", PCI->irq);

	{
		int i;
		spci_info_t *psi;
		
		psi = (spci_info_t *)pci_get_drvdata (PCI);
		for (i=0; i<3; i++)
			PRINTK("memio_phy%d=0x%08lx\n", i, psi->mmio_base[i]);
	}
	
	return 0;
}

static int spci_release( struct inode *inode, struct file *filp)
{
	PRINTK("\n");

	if ( PCI == NULL )
	{
		PRINTK("not found PCI resource\n");
		return -ENOMEM;
	}

	{
		spci_info_t *psi;
		psi = (spci_info_t *)pci_get_drvdata (PCI);
		dma_free_coherent(NULL, TXDMA_SIZE, psi->tx_vir, psi->txdma_phy);
	}

	free_irq( PCI->irq, PCI );
	return 0;
}

static ssize_t spci_read(struct file *file, char *buf, size_t count, loff_t *ptr)
{
    spci_info_t *psi = (spci_info_t *)pci_get_drvdata (PCI);
    memcpy(buf, psi->mmio_vir[1] + psi->offset[1], count);
    return count;
}

int dma_start(int channel, int mode, unsigned char *dma_buff, unsigned int count)
{
	unsigned long flags;
	
	flags = claim_dma_lock();
	disable_dma(channel);
	clear_dma_ff(channel);
	set_dma_mode(channel, mode);
	set_dma_addr(channel, virt_to_bus(dma_buff));
	set_dma_count(channel, count);
	enable_dma(channel);
	release_dma_lock(flags);
	
	return 0;
}

static ssize_t spci_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
    spci_info_t *psi = (spci_info_t *)pci_get_drvdata (PCI);

	PRINTK("addr=%08x, ch=%d\n, sz=%d", (int)(psi->mmio_vir[0]+psi->offset[0]), psi->dma_ch, count);

	if (count > 1)
	{	
		copy_from_user(psi->tx_vir, buf, count);
		dma_start(psi->dma_ch, DMA_MODE_WRITE, (unsigned char *)psi->mmio_vir[0], count); 
	}
	else
	{
		memcpy(psi->mmio_vir[0] + psi->offset[0], buf, count);
	}

    return count;
}

int spci_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    spci_info_t *psi = (spci_info_t *)pci_get_drvdata (PCI);

	int *data = (int *)arg;

	switch (cmd)
	{
		case WCMD_WOFFSET:
			psi->offset[0] = *data;
			break;

		case WCMD_ROFFSET:
			psi->offset[1] = *data;
			break;

		case RCMD_WOFFSET:
			return psi->offset[0];

		case RCMD_ROFFSET:
			return psi->offset[1];
	}
	PRINTK("cmd=%d, data=%08x\n", cmd, *data);
    return 0;
}

static struct file_operations spci_fops =
{
	open    : spci_open, 
	release : spci_release, 
	read    : spci_read,
	write   : spci_write,
	ioctl   : spci_ioctl,
};

static void __devexit spci_remove (struct pci_dev *pdev)
{
	spci_info_t *psi = pci_get_drvdata (pdev);

	iounmap( psi->mmio_vir );
	
	pci_release_regions (pdev);

	unregister_chrdev( major, DEV_NAME );

	if (psi) kfree( psi);
		
	pci_set_drvdata (pdev, NULL);
	PRINTK("\n");
}

static int __devinit spci_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	spci_info_t *psi;

	int i, rc;
	int err, pci_using_dac;
	unsigned long mmio_start[3], mmio_end[3], mmio_flags[3], mmio_len[3];
	
	
	{
		unsigned char  pci_rev;
		
		pci_read_config_byte (pdev, PCI_REVISION_ID, &pci_rev);
		PRINTK( "pci_rev=%d\n", pci_rev );
	}

	if (pci_enable_device (pdev) < 0)
		return -EIO;

	for (i=0; i<3; i++)
	{
		mmio_start[i]  = pci_resource_start (pdev, i);
		mmio_end[i]    = pci_resource_end   (pdev, i);
		mmio_flags[i]  = pci_resource_flags (pdev, i);	// IORESOURCE
		mmio_len[i]    = pci_resource_len   (pdev, i);
		PRINTK("MMIO%d  base=0x%lx len=0x%lx\n", i, mmio_start[i], mmio_len[i]);
	}

	rc = pci_request_regions (pdev, DEV_NAME);
	if (rc)
	{
		PRINTK( "pci_request_regions error\n" );
		return -EBUSY;
	}

	pci_set_master (pdev);
	
	psi = (spci_info_t *)kmalloc( sizeof(spci_info_t), GFP_KERNEL );
	memset( psi, 0 , sizeof(spci_info_t) );

	psi->mmio_base[0] = mmio_start[0];
	psi->mmio_base[1] = mmio_start[1];
	psi->mmio_base[2] = mmio_start[2];

	psi->mmio_vir[0]  = pci_iomap(pdev, 0, mmio_len[0]);
	psi->mmio_vir[1]  = pci_iomap(pdev, 1, mmio_len[1]);
	psi->mmio_vir[2]  = pci_iomap(pdev, 2, mmio_len[2]);
	PRINTK( "remap MMIO0 =0x%p\n", psi->mmio_vir[0] );
	PRINTK( "remap MMIO1 =0x%p\n", psi->mmio_vir[1] );
	PRINTK( "remap MMIO2 =0x%p\n", psi->mmio_vir[2] );
	
	pci_set_drvdata (pdev, psi);

	major &= 0xff;
	if( register_chrdev( major, DEV_NAME, &spci_fops ) )  
	{        
		PRINTK("unable to get major %d for %s \n", major, DEV_NAME );
		// return -EBUSY;	
	}
	else
	{
		PRINTK("major %d for %s \n", major, DEV_NAME );
	}
	PCI = pdev;

	// DMA Setup
	if (!(err = pci_set_dma_mask(pdev, DMA_BIT_MASK(64)))) {
		pci_using_dac = 1;
	} 
	else if (!(err = pci_set_dma_mask(pdev, DMA_BIT_MASK(32)))) {
		pci_using_dac = 0;
	}
	else
	{
		PRINTK("No usable DMA configuration, abouting\n");
		return err;
	}
	PRINTK("pci_using_dac = %d\n", pci_using_dac);

	// DMA Memory allocation
	psi->tx_vir = dma_alloc_coherent(NULL, TXDMA_SIZE, &psi->txdma_phy, GFP_KERNEL);
	if (psi->tx_vir == NULL)
		PRINTK("Tx DMA Memory alloc failed!!!\n");
	PRINTK("tx_vir: %08x, txdma_phy: %08x\n", (int)psi->tx_vir, (int)psi->txdma_phy);

	for (i=0; i<10; i++)
		if (request_dma(i, DEV_NAME) == 0)
			break;
	psi->dma_ch = i;
	PRINTK("DMA Channel Number=%d\n", i);

	return 0;
}

static struct pci_driver spci_driver = {
	.name		= DEV_NAME,
	.id_table	= spci_tbl,
	.probe		= spci_probe,
	.remove		= spci_remove,
};

static int __init spci_init( void )
{
	int	rtn;
	
	rtn = pci_register_driver(&spci_driver);

	return rtn;
}
static void __exit	spci_exit( void )
{
	pci_unregister_driver(&spci_driver);
}

/*-----------------------------------------------------------------------------*/

module_init(spci_init);
module_exit(spci_exit);

MODULE_PARM_DESC(major , "a");
MODULE_AUTHOR("test");
MODULE_LICENSE("Dual BSD/GPL");

/* end */



